from .service import PlazaService
from .service_configuration import ServiceConfiguration
from .blocks import (
    ServiceBlock,
    ServiceTriggerBlock,
    BlockArgument,
    DynamicBlockArgument,
    BlockType,
    VariableBlockArgument,
    BlockContext,
    VariableClass,
)
from .registration import MessageBasedServiceRegistration, FormBasedServiceRegistration



# Server -> Bridge
CALL_MESSAGE_TYPE = "FUNCTION_CALL"
GET_HOW_TO_SERVICE_REGISTRATION = "GET_HOW_TO_SERVICE_REGISTRATION"
REGISTRATION_MESSAGE = "REGISTRATION"
OAUTH_RETURN = "OAUTH_RETURN"
DATA_CALLBACK = "CALLBACK"

# Bridge -> Server
AUTHENTICATION = "AUTHENTICATION"
NOTIFICATION = "NOTIFICATION"
CONFIGURATION = "CONFIGURATION"



class ServiceConfiguration:
    def __init__(self, service_name, blocks, registration=None, is_public=False):
        self.service_name = service_name
        self.blocks = blocks
        self.is_public = is_public
        self.registration = registration

    def serialize(self):
        serialized_registration = None
        if self.registration is not None:
            serialized_registration = self.registration.serialize()

        return {
            "service_name": self.service_name,
            "blocks": list(map(lambda b: b.serialize(), self.blocks)),
            "is_public": self.is_public,
            "registration": serialized_registration,
        }



import asyncio
import json
import logging
import time
import traceback

import websockets

from . import protocol
from .service_configuration import ServiceConfiguration

SLEEP_BETWEEN_RETRIES = 5


class ExtraData:
    def __init__(self, user_id, extra_data_dict=None):
        self.user_id = user_id
        if extra_data_dict is None:
            self.extra_data_dict = {}
        else:
            self.extra_data_dict = extra_data_dict

        self._last_monitor_value = self.extra_data_dict.get("last_monitor_value", None)
        if self._last_monitor_value is None:
            self.last_monitor_value = None
        else:
            self.last_monitor_value = self._last_monitor_value.get("value", None)


class AnswerHandler:
    def __init__(self, message_id, websocket):
        self.message_id = message_id
        self.websocket = websocket


class PlazaService:
    def __init__(self, service_url, service_token):
        self.service_url = service_url
        self.service_token = service_token
        self.INTERNAL_FUNCTION_NAMES = {"__ping": self.__answer_ping}
        self.__registerer = None
        self.loop = None
        self.websocket = None

    async def __answer_ping(self, websocket, message):
        (_msg_type, _value, message_id) = message
        await websocket.send(
            json.dumps({"message_id": message_id, "success": True, "result": "PONG"})
        )

    async def emit_event(self, to_user, key, content=None, event=None):
        await self.websocket.send(
            json.dumps(
                {
                    "type": protocol.NOTIFICATION,
                    "key": key,
                    "to_user": to_user,
                    "value": event,
                    "content": content,
                }
            )
        )

    def emit_event_sync(self, to_user, key, content=None, event=None):
        result = asyncio.run_coroutine_threadsafe(
            self.emit_event(to_user, key, event=event, content=content), self.loop
        )
        return result

    def __parse(self, message):
        parsed = json.loads(message)
        return (
            parsed["type"],
            parsed["value"],
            parsed["message_id"],
            ExtraData(parsed.get("user_id"), parsed.get("extra_data", None)),
        )

    async def __interact(self, websocket):
        async for message in websocket:
            logging.debug("Received: {}".format(message))
            (msg_type, value, message_id, extra_data) = self.__parse(message)

            if msg_type == protocol.CALL_MESSAGE_TYPE:
                function_name = value["function_name"]

                if function_name in self.INTERNAL_FUNCTION_NAMES:
                    await self.INTERNAL_FUNCTION_NAMES[function_name](
                        websocket, (msg_type, value, message_id), extra_data
                    )
                else:
                    try:
                        response = await self.handle_call(
                            function_name, value["arguments"], extra_data
                        )
                    except:
                        logging.warn(traceback.format_exc())
                        await websocket.send(
                            json.dumps({"message_id": message_id, "success": False})
                        )
                        continue

                    await websocket.send(
                        json.dumps(
                            {
                                "message_id": message_id,
                                "success": True,
                                "result": response,
                            }
                        )
                    )

            elif msg_type == protocol.GET_HOW_TO_SERVICE_REGISTRATION:
                if self.__registerer is None:
                    await websocket.send(
                        json.dumps(
                            {"message_id": message_id, "success": True, "result": None}
                        )
                    )
                else:
                    await websocket.send(
                        json.dumps(
                            {
                                "message_id": message_id,
                                "success": True,
                                "result": self.__registerer.serialize(extra_data),
                            }
                        )
                    )

            elif msg_type == protocol.REGISTRATION_MESSAGE:
                if self.__registerer is None:
                    await websocket.send(
                        json.dumps(
                            {
                                "message_id": message_id,
                                "success": False,
                                "error": "No registerer available",
                            }
                        )
                    )
                else:
                    result = await self.__registerer.register(value, extra_data)
                    message = None
                    if result != True:
                        result, message = result

                    await websocket.send(
                        json.dumps(
                            {
                                "message_id": message_id,
                                "success": result,
                                "message": message,
                            }
                        )
                    )

            elif msg_type == protocol.OAUTH_RETURN:
                if self.__registerer is None:
                    await websocket.send(
                        json.dumps(
                            {
                                "message_id": message_id,
                                "success": False,
                                "error": "No registerer available",
                            }
                        )
                    )
                else:
                    result = await self.__registerer.register(value, extra_data)
                    message = None
                    if result != True:
                        result, message = result

                    await websocket.send(
                        json.dumps(
                            {
                                "message_id": message_id,
                                "success": result,
                                "message": message,
                            }
                        )
                    )

            elif msg_type == protocol.DATA_CALLBACK:
                try:
                    response = await self.handle_data_callback(
                        value["callback"], extra_data
                    )
                except:
                    logging.warn(traceback.format_exc())
                    await websocket.send(
                        json.dumps({"message_id": message_id, "success": False})
                    )
                    continue

                await websocket.send(
                    json.dumps(
                        {"message_id": message_id, "success": True, "result": response}
                    )
                )

            else:
                raise Exception("Unknown message type “{}”".format(msg_type))

    async def __connect(self):
        async with websockets.connect(self.service_url) as websocket:
            self.websocket = websocket

            await websocket.send(
                json.dumps(
                    {
                        "type": protocol.AUTHENTICATION,
                        "value": {"token": self.service_token},
                    }
                )
            )

            logging.info("Connected successfully")
            configuration = self.handle_configuration()
            self.__registerer = configuration.registration
            if not isinstance(configuration, ServiceConfiguration):
                raise TypeError(
                    "Configuration has to inherit plaza_service.ServiceConfiguration"
                )

            logging.debug(
                "Setting configuration: {}".format(
                    json.dumps(configuration.serialize(), indent=4)
                )
            )

            await websocket.send(
                json.dumps(
                    {"type": protocol.CONFIGURATION, "value": configuration.serialize()}
                )
            )

            await self.__interact(websocket)

    def run(self):
        self.loop = asyncio.get_event_loop()
        while True:
            try:
                asyncio.get_event_loop().run_until_complete(self.__connect())
            except KeyboardInterrupt:
                return
            except:
                logging.warn(traceback.format_exc())

            logging.debug("Waiting {}s for reconnection".format(SLEEP_BETWEEN_RETRIES))
            time.sleep(SLEEP_BETWEEN_RETRIES)
            logging.debug("Reconnecting")



import enum


class BlockContextArgument:
    def __getitem__(self, index):
        return {"type": "argument", "index": index}


class BlockContextClass:
    def __init__(self):
        self.ARGUMENTS = BlockContextArgument()


BlockContext = BlockContextClass()


class BlockType(enum.Enum):
    TRIGGER = "trigger"
    OPERATION = "operation"
    GETTER = "getter"


class ServiceBlock:
    def __init__(
        self,
        id,
        function_name,
        message,
        block_type,
        block_result_type=None,
        arguments=[],
        save_to=None,
    ):
        self.id = id
        self.function_name = function_name
        self.message = message
        self.block_type = block_type
        self.block_result_type = block_result_type
        self.arguments = arguments
        self.save_to = save_to

    def serialize(self):
        return {
            "id": self.id,
            "function_name": self.function_name,
            "message": self.message,
            "arguments": list(map(lambda a: a.serialize(), self.arguments)),
            "block_type": self.block_type.value,
            "block_result_type": self.block_result_type,
            "save_to": self.save_to,
        }


class ServiceTriggerBlock:
    def __init__(
        self,
        id,
        function_name,
        message,
        arguments=[],
        save_to=None,
        expected_value=None,
        key=None,
    ):
        self.id = id
        self.function_name = function_name
        self.message = message
        self.arguments = arguments
        self.save_to = save_to
        self.expected_value = expected_value
        if key is not None:
            self.key = key
        else:
            self.key = function_name

    def serialize(self):
        return {
            "id": self.id,
            "function_name": self.function_name,
            "message": self.message,
            "arguments": list(map(lambda a: a.serialize(), self.arguments)),
            "save_to": self.save_to,
            "expected_value": self.expected_value,
            "block_type": BlockType.TRIGGER.value,
            "key": self.key,
        }


ALLOWED_ARGUMENT_TYPES = {
    str: "string",
    int: "integer",
    float: "float",
    bool: "boolean",
}


class VariableClass(enum.Enum):
    SINGLE = "single"  # Like atoms in lisp
    LIST = "list"


class BlockArgument:
    def __init__(self, type, default):
        if type not in ALLOWED_ARGUMENT_TYPES:
            raise TypeError("Type “{}” not allowed".format(type))

        self.type = type
        self.default = default

    def serialize(self):
        return {"type": ALLOWED_ARGUMENT_TYPES[self.type], "default": self.default}


class VariableBlockArgument:
    def __init__(self, _class=None):
        if _class is not None:
            assert isinstance(_class, VariableClass)
            self.variable_class = _class.value
        else:
            self.variable_class = VariableClass.SINGLE.value

    def serialize(self):
        return {"type": "variable", "class": self.variable_class}


class DynamicBlockArgument:
    def __init__(self, type, callback):
        if type not in ALLOWED_ARGUMENT_TYPES:
            raise TypeError("Type “{}” not allowed".format(type))

        self.type = type
        self.callback = callback

    def serialize(self):
        return {
            "type": ALLOWED_ARGUMENT_TYPES[self.type],
            "values": {"callback": self.callback},
        }



from setuptools import setup

setup(
    name="plaza_service",
    version="0.0.2",
    description="Helper to build plaza services.",
    author="kenkeiras",
    author_email="kenkeiras@codigoparallevar.com",
    url="https://gitlab.com/plaza-project/bridges/python-plaza-lib",
    license="Apache License 2.0",
    packages=["plaza_service"],
    scripts=[],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Development Status :: 1 - Planning",
        "Intended Audience :: Developers",
    ],
    include_package_data=True,
    install_requires=["websockets"],
    zip_safe=False,
)
